----------------------------------------------------------------------------------
-- Company: 
-- Engineer: Mlinkó Péter, Vági Levente
-- 
-- Create Date: 12/10/2018 01:49:49 PM
-- Design Name: 
-- Module Name: game_of_life - Behavioral
-- Project Name: FPGA grand project
-- Target Devices: 
-- Tool Versions: 
-- Description: 
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity game_of_life is
    generic (cell_number : positive := 32);
    Port ( clk : in STD_LOGIC;                             -- 100MHz clock sign
           switch_mode : in STD_LOGIC;                     -- 0: editor, 1: simulation
           move_left : in STD_LOGIC;                       -- cursor's move left button
           move_right : in STD_LOGIC;                      -- cursor's move right button
           move_up : in STD_LOGIC;                         -- cursor's move up button
           move_down : in STD_LOGIC;                       -- cursor's move down button
           set_cell : in STD_LOGIC;                        -- invert cell's value
           
           Hsync : out STD_LOGIC;                          -- horizontal synchronization controll sign
           Vsync : out STD_LOGIC;                          -- vertical synchronization controll sign
           frame : out STD_LOGIC_VECTOR(11 downto 0)       -- RGB code of actual pixel point
    );
end game_of_life;

architecture Behavioral of game_of_life is

component clk_wiz_0
port
 (-- Clock in ports
  -- Clock out ports
  clk_out1          : out    std_logic;
  -- Status and control signals
  locked            : out    std_logic;
  clk_in1           : in     std_logic
 );
end component;

signal clk_25MHz, locked : std_logic;
signal cells : std_logic_vector(0 to cell_number*cell_number-1) := (501 => '1', 502 => '1', 503 => '1',others => '0');
signal cells_to_draw : std_logic_vector(0 to cell_number*cell_number-1) := (501 => '1', 502 => '1', 503 => '1',others => '0');

-- drawing period's constants
constant HORIZONTAL_SYNC_PULSE : integer := 800;
constant VERTICAL_SYNC_PULSE : integer := 521;
constant HORIZONTAL_DISPLAY_TIME : integer := 640;
constant VERTICAL_DISPLAY_TIME : integer := 480;
constant HORIZONTAL_PULSE_WIDTH : integer := 96;
constant VERTICAL_PULSE_WIDTH : integer := 2;
constant HORIZONTAL_FRONT_PORCH : integer := 16;
constant VERTICAL_FRONT_PORCH : integer := 10;
constant HORIZONTAL_BACK_PORCH : integer := 48;
constant VERTICAL_BACK_PORCH : integer := 29;

-- helpers for drawing
signal horizontal_counter, vertical_counter : unsigned(9 downto 0);
signal Hsync_sig, Vsync_sig : std_logic;
signal frame_sig : std_logic_vector(11 downto 0);

-- simulation processers
signal shift_register_top : std_logic_vector(0 to cell_number+1);
signal shift_register_middle : std_logic_vector(0 to cell_number+1);
signal actual_register : std_logic_vector(0 to 0);

signal neighbours : std_logic_vector(0 to 8);
signal neighbours_count : unsigned(3 downto 0);
signal frame_counter : unsigned(6 downto 0);

-- cursor handling
signal cursor_pos_x, cursor_pos_y : unsigned(9 downto 0);

signal prev_move_left : std_logic;
signal prev_move_right : std_logic;
signal prev_move_up : std_logic;
signal prev_move_down : std_logic;
signal prev_set_cell : std_logic;

begin

clk_gen_25MHz : clk_wiz_0
   port map ( 
  -- Clock out ports  
   clk_out1 => clk_25MHz,
  -- Status and control signals                
   locked => locked,
   -- Clock in ports
   clk_in1 => clk
 );

options : process(clk_25MHz,locked)
begin
    if locked = '0' then
        horizontal_counter <= (others => '0');
        vertical_counter <= (others => '0');
        Hsync_sig <= '1';
        Vsync_sig <= '1';
        
        shift_register_top <= (others => '0');
        shift_register_middle <= (others => '0');
        actual_register <= (others => '0');
        neighbours <= (others => '0');
        neighbours_count <= (others => '0');
        frame_counter <= (others => '0');
        
        cursor_pos_x <= (others => '0');
        cursor_pos_y <= (others => '0');
        prev_move_left <= move_left;
        prev_move_right <= move_right;
        prev_move_up <= move_up;
        prev_move_down <= move_down;
        prev_set_cell <= set_cell;
    else
        if clk_25MHz'event and clk_25MHz = '1' then
            -- output change
            if horizontal_counter(9 downto 3) < cell_number and vertical_counter(9 downto 3) < cell_number then
                -- editor mode
                if switch_mode = '0' then
                    if cursor_pos_x <= horizontal_counter and cursor_pos_x+8 > horizontal_counter and cursor_pos_y <= vertical_counter and cursor_pos_y+8 > vertical_counter then
                        if cells_to_draw(to_integer(vertical_counter(9 downto 3) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3))) = '0' then
                            frame_sig <= "000011110000";
                        else
                            frame_sig <= "111100000000";
                        end if;
                    else
                        frame_sig <= (others => cells_to_draw(to_integer(vertical_counter(9 downto 3) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3))));
                    end if;

                -- simulation mode
                else
                    frame_sig <= (others => cells_to_draw(to_integer(vertical_counter(9 downto 3) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3))));
                end if;
            else
                frame_sig <= (others => '0');
            end if;

            -- handle shift registers
            if switch_mode = '1' and frame_counter = 0 and horizontal_counter(2 downto 0) = 0 and horizontal_counter(9 downto 3) < cell_number+2 and vertical_counter(2 downto 0) = 0 and vertical_counter(9 downto 3) < cell_number+2 then
                shift_register_top <= shift_register_top(1 to cell_number+1) & shift_register_middle(0);
                
                if horizontal_counter(9 downto 3) /= 0 and horizontal_counter(9 downto 3) < cell_number+1 and vertical_counter(9 downto 3) /= 0 and vertical_counter(9 downto 3) < cell_number+1 then
                    shift_register_middle <= shift_register_middle(1 to cell_number+1) & actual_register;
                    neighbours <= neighbours(3 to 8) & shift_register_top(0) & shift_register_middle(0) & actual_register;
                    neighbours_count <= neighbours_count - unsigned(neighbours(0 to 0)) - unsigned(neighbours(1 to 1)) - unsigned(neighbours(2 to 2)) + unsigned(shift_register_top(0 to 0)) + unsigned(shift_register_middle(0 to 0)) + unsigned(actual_register);
                    actual_register(0) <= cells(to_integer((vertical_counter(9 downto 3)-1)* to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-1));
                else
                    shift_register_middle <= shift_register_middle(1 to cell_number+1) & actual_register;
                    neighbours <= neighbours(3 to 8) & shift_register_top(0) & shift_register_middle(0) & actual_register;
                    
                    
                    neighbours_count <= neighbours_count - unsigned(neighbours(0 to 0)) - unsigned(neighbours(1 to 1)) - unsigned(neighbours(2 to 2)) + unsigned(shift_register_top(0 to 0)) + unsigned(shift_register_middle(0 to 0)) + unsigned(actual_register);
                    actual_register(0) <= '0';
                end if;
                
                if vertical_counter(9 downto 3) >= 2 and horizontal_counter(9 downto 3) >= 2 then
                    
                    if horizontal_counter(9 downto 3) = 0 or horizontal_counter(9 downto 3) = cell_number+1 or vertical_counter(9 downto 3) = 0 or vertical_counter(9 downto 3) = cell_number+1 then
                        cells(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                        cells_to_draw(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                    else
                        if neighbours(4)='1' then
                            if neighbours_count = 3 or neighbours_count = 4 then
                                cells(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '1';
                                cells_to_draw(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '1';
                            else
                                cells(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                                cells_to_draw(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                            end if;
                        else
                            if neighbours_count = 3 then
                                cells(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '1';
                                cells_to_draw(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '1';
                            else
                                cells(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                                cells_to_draw(to_integer((vertical_counter(9 downto 3)-2) * to_unsigned(cell_number,8) + horizontal_counter(9 downto 3)-4)) <= '0';
                            end if;
                        end if;
                    end if;
                    
                end if;
                
            end if;
            
            -- moving with cursor
            if switch_mode = '0' then
                if prev_move_left = '0' and move_left = '1' and cursor_pos_x /= 0 then
                    cursor_pos_x <= cursor_pos_x - 8;
                end if;
                if prev_move_right = '0' and move_right = '1' and cursor_pos_x /= (cell_number-1)*8 then
                    cursor_pos_x <= cursor_pos_x + 8;
                end if;
                if prev_move_up = '0' and move_up = '1' and cursor_pos_y /= 0 then
                    cursor_pos_y <= cursor_pos_y - 8;
                end if;
                if prev_move_down = '0' and move_down = '1' and cursor_pos_y /= (cell_number-1)*8 then
                    cursor_pos_y <= cursor_pos_y + 8;
                end if;
                
                -- set cell's value
                if prev_set_cell = '0' and set_cell = '1' then
                    cells(to_integer(cursor_pos_y(9 downto 3)*cell_number + cursor_pos_x(9 downto 3))) <= not cells(to_integer(cursor_pos_y(9 downto 3)*cell_number + cursor_pos_x(9 downto 3)));
                    cells_to_draw(to_integer(cursor_pos_y(9 downto 3)*cell_number + cursor_pos_x(9 downto 3))) <= not cells_to_draw(to_integer(cursor_pos_y(9 downto 3)*cell_number + cursor_pos_x(9 downto 3)));
                end if;
            end if;
            
            -- memorize previous state
            prev_move_left <= move_left;
            prev_move_right <= move_right;
            prev_move_up <= move_up;
            prev_move_down <= move_down;
            prev_set_cell <= set_cell;
            
            -- sign options
            if horizontal_counter = HORIZONTAL_DISPLAY_TIME + HORIZONTAL_FRONT_PORCH - 1 then
                Hsync_sig <= '0';
            end if;
            if horizontal_counter = HORIZONTAL_DISPLAY_TIME + HORIZONTAL_FRONT_PORCH + HORIZONTAL_PULSE_WIDTH - 1 then
                Hsync_sig <= '1';
            end if;
            if vertical_counter = VERTICAL_DISPLAY_TIME + VERTICAL_FRONT_PORCH - 1 then
                Vsync_sig <= '0';
            end if;
            if vertical_counter = VERTICAL_DISPLAY_TIME + VERTICAL_FRONT_PORCH + VERTICAL_PULSE_WIDTH - 1 then
                Vsync_sig <= '1';
            end if;
            
            -- position changing
            if vertical_counter = VERTICAL_SYNC_PULSE-1 then
                vertical_counter <= (others => '0');
                neighbours_count <= (others => '0');
                neighbours <= (others => '0');
                if frame_counter = 63 then
                    frame_counter <= (others => '0');
                else
                    frame_counter <= frame_counter + 1;
                end if;
            end if;
            if horizontal_counter = HORIZONTAL_SYNC_PULSE-1 then
                horizontal_counter <= (others => '0');
                vertical_counter <= vertical_counter + 1;
            else
                horizontal_counter <= horizontal_counter + 1;
            end if;
        end if;
    end if;
end process;

draw : process(clk_25MHz)
begin
    if clk_25MHz'event and clk_25MHz = '1' then
        Hsync <= Hsync_sig;
        Vsync <= Vsync_sig;
        frame <= frame_sig;
    end if;
end process;

end Behavioral;
